'use strict';

var url = require('url');
var oAuth1Authenticate = require('./oauth1');
var oAuth2Authenticate = require('./oauth2');

var SUPPORTED_GRANTS = ['token', 'code', 'credentials'];
var RAML_1_SUPPORTED_GRANTS = ['implicit', 'authorization_code', 'client_credentials'];
var compatibilityMapping = {
  implicit: 'token',
  authorization_code: 'code',
  client_credentials: 'credentials'
};

function authenticate(options) {
  if (options.type === 'Basic Authentication') {
    if (typeof options.username !== 'string' || typeof options.password !== 'string') {
      return Promise.reject(new TypeError('Username and password must be defined'));
    }

    return Promise.resolve({ username: options.username, password: options.password });
  }

  var oauthCallback = options.oauthCallback || '/authentication/oauth.html';
  var REDIRECT_URI = global.location ? url.resolve(global.location.href, oauthCallback) : null;

  options = Object.assign({
    redirectUri: REDIRECT_URI
  }, options);

  if (options.type === 'OAuth 1.0') {
    return oAuth1Authenticate(options);
  }

  if (options.type === 'OAuth 2.0') {
    // Sanitize authorization grants to an array.
    if (typeof options.authorizationGrants === 'string') {
      options.authorizationGrants = [options.authorizationGrants];
    }

    // Use insection to get the accepted grant types in the order of the
    // supported grant types (which are ordered by preference).
    var grantType = SUPPORTED_GRANTS.find(function (grant) {
      return options.authorizationGrants.some(function (authorizationGrant) {
        return authorizationGrant === grant;
      });
    });

    if (!grantType) {
      grantType = RAML_1_SUPPORTED_GRANTS.find(function (grant) {
        return options.authorizationGrants.some(function (authorizationGrant) {
          return authorizationGrant === grant;
        });
      });
      // eslint-disable-next-line prefer-destructuring
      grantType = compatibilityMapping[grantType];
    }

    if (!grantType) {
      return Promise.reject(new Error('Unsupported OAuth2 Grant Flow. Supported flows include ' + RAML_1_SUPPORTED_GRANTS.join(', ')));
    }

    return oAuth2Authenticate[grantType](options);
  }

  return Promise.reject(new Error('Authentication flow not supported'));
}

module.exports = authenticate;
//# sourceMappingURL=index.js.map