'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var authenticate = require('./authenticate');
var clientGenerator = require('./generator');
var fromPath = require('./fromPath');

var _require = require('./client'),
    resolveClient = _require.resolveClient;

var CONFIG_PROPERTY = '!config';
var CLIENT_PROPERTY = '!client';
var isBrowser = !!process.browser;
var defaultQueryBuilder = function defaultQueryBuilder(_ref) {
  var groupId = _ref.groupId,
      assetId = _ref.assetId,
      version = _ref.version;
  return 'https://anypoint.mulesoft.com/exchange/api/v1/assets/' + groupId + '/' + assetId + '/' + version + '/api/json';
};

module.exports = function clientCreator() {
  var ramlClientOptions = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var RAMLClient = {};
  var contextWindow = ramlClientOptions.contextWindow || global;
  var editor = ramlClientOptions.editor;

  var clientId = 0;

  /**
   * Responsible for loading RAML documents and return API clients.
   *
   * @param {String}   name
   * @param {String}   uri
   * @param {Object}   [clientConfig]
   * @param {Function} done
   */

  RAMLClient.createClient = function () {
    var name = void 0;
    var source = void 0;
    var clientConfig = void 0;
    var groupId = void 0;
    var assetId = void 0;
    var version = void 0;
    var addToContext = false;

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (args.length <= 3) {
      if (_typeof(args[0]) === 'object') {
        source = args[0];
        clientConfig = args[1];
      } else if (typeof args[0] === 'string' && typeof args[1] === 'string') {
        name = args[0];
        source = args[1];
        clientConfig = args[2];
      } else {
        source = args[0];
        clientConfig = args[1];
      }

      if (name) {
        addToContext = true;
      }
    } else {
      if (!ramlClientOptions.ramlQueryBuilder) {
        ramlClientOptions.ramlQueryBuilder = defaultQueryBuilder;
      }

      if (args.slice(0, 4).every(function (arg) {
        return typeof arg === 'string';
      })) {
        name = args[0];
        groupId = args[1];
        assetId = args[2];
        version = args[3];
        clientConfig = args[4];

        addToContext = true;
      } else if (args.slice(0, 3).every(function (arg) {
        return typeof arg === 'string';
      })) {
        groupId = args[0];
        assetId = args[1];
        version = args[2];
        clientConfig = args[3];
      } else {
        return Promise.reject(new Error('Invalid arguments'));
      }
    }

    if (!name) {
      name = 'client' + (clientId || '');
      clientId += 1;
    }

    clientConfig = clientConfig || {};

    /**
     * Generate and attach the RAML client from RAML.
     *
     * @param  {Object} raml
     */
    var createClient = function createClient(raml) {
      if (isBrowser) {
        var client = clientGenerator(name, raml, Object.assign({}, ramlClientOptions, clientConfig));

        if (addToContext) {
          contextWindow[name] = client;
        }

        return fromPath({}, name.split('.'), client);
      }

      if ((typeof raml === 'undefined' ? 'undefined' : _typeof(raml)) === 'object') {
        return clientGenerator(name, raml, Object.assign({}, ramlClientOptions, clientConfig));
      }

      return Promise.reject(new Error('Invalid RAML'));
    };

    var resolveClientFn = ramlClientOptions.ramlResolver || resolveClient;

    return resolveClientFn({
      groupId: groupId, assetId: assetId, version: version, url: source, options: ramlClientOptions
    }).then(createClient);
  };

  /**
   * Set a configuration value on a client.
   *
   * @param {Function} client
   * @param {String}   key
   * @param {*}        value
   */
  RAMLClient.set = function (client, key, value) {
    // If we don't have enough arguments for a key and value, assume we have
    // a fresh configuration object.
    if (arguments.length < 3) {
      return Object.assign(client[CONFIG_PROPERTY], key);
    }

    client[CONFIG_PROPERTY][key] = value;

    return value;
  };

  /**
   * Retrieve a value from the client config object.
   *
   * @param  {Function} client
   * @param  {String}   key
   * @return {*}
   */
  RAMLClient.get = function (client, key) {
    if (arguments.length < 2) {
      return client[CONFIG_PROPERTY];
    }

    return client[CONFIG_PROPERTY][key];
  };

  /**
   * Unset a key from the client configuration.
   *
   * @param  {Function} client
   * @param  {String}   key
   * @return {Boolean}
   */
  RAMLClient.unset = function (client, key) {
    if (arguments.length < 2) {
      Object.keys(client[CONFIG_PROPERTY]).forEach(function ($key) {
        delete client[CONFIG_PROPERTY][$key];
      });

      return true;
    }

    return delete client[CONFIG_PROPERTY][key];
  };

  /**
   * Authenticate a RAML API client passing an optional method and accompanying
   * options object.
   *
   * @param {Function} client
   * @param {String}   method
   * @param {Object}   options
   * @param {Function} done
   */
  RAMLClient.authenticate = function (client, method, authOptions) {
    var _this = this;

    var clientOption = client[CLIENT_PROPERTY];
    var securitySchemes = clientOption.securitySchemes;

    // Possibility to use two params

    if (method && (typeof method === 'undefined' ? 'undefined' : _typeof(method)) === 'object' && !authOptions) {
      authOptions = method;
      method = null;
    }

    // Extend passed in options with global options (for `baseUriParameters`).
    authOptions = Object.assign({}, authOptions, client[CONFIG_PROPERTY], ramlClientOptions);

    return authenticate(securitySchemes, method, authOptions).then(function (_ref2) {
      var scheme = _ref2.scheme,
          options = _ref2.options,
          tokens = _ref2.tokens;

      // Alias onto the client object for future use.
      clientOption.authentication[scheme.type] = Object.assign({}, options, tokens);

      // Set private fields to asterisks of an arbitrary length.
      Object.keys(tokens).forEach(function (key) {
        if (authenticate.SECRET_FIELDS[key]) {
          tokens[key] = new Array(Math.ceil(tokens[key].length / 5) * 5).join('*');
        }
      });

      if (isBrowser) {
        return tokens;
      }

      _this.set(client, 'session', tokens);

      return client;
    });
  };

  if (editor && !editor.languages.typescript.javascriptDefaults.getExtraLibs()['api.d.ts']) {
    editor.languages.typescript.javascriptDefaults.addExtraLib('\n      declare class API {\n        /**\n         * Generate an API client from a RAML document.\n         *\n         * @param {string} groupId\n         * @param {string} assetId\n         * @param {string} version\n         * @param {any} config\n         */\n        static createClient(groupId: string, assetId: string, version: string, config?: { sync: boolean, baseUri: string }): any;\n\n        /**\n         * Generate an API client from a RAML document and alias it on the window.\n         *\n         * @param {string} name\n         * @param {string} groupId\n         * @param {string} assetId\n         * @param {string} version\n         * @param {any} config\n         */\n        static createClient(name: string, groupId: string, assetId: string, version: string, config?: { sync: boolean, baseUri: string }): any;\n\n        /**\n         * Generate an API client from a RAML document and alias it on the window.\n         *\n         * @param {string} url\n         * @param {any} config\n         */\n        static createClient(name: string, url: string, config?: { sync: boolean, baseUri: string }): any;\n\n        /**\n         * Set a configuration option of a RAML API client.\n         *\n         * @param {any} client\n         * @param {string} key\n         * @param {any} value\n         */\n        static set(client: any, key: string, value: any): any;\n\n        /**\n         * Get a configuration option from a RAML API client.\n         *\n         * @param {any} client\n         * @param {string} key\n         */\n        static get(client: any, key: string): any;\n\n        /**\n         * Authenticate a RAML API client with credentials.\n         *\n         * @param {any} client\n         * @param {string} method\n         * @param {any} options\n         */\n        static authenticate(client: any, method: string, options?: any): any;\n      }\n    '.replace(/ {6}/g, ''), 'api.d.ts');

    editor.languages.typescript.javascriptDefaults.addExtraLib('\n      declare class APIResponse {\n        body: any;\n        status: number;\n        headers: any;\n      }\n    '.replace(/ {6}/g, ''), 'api-response.d.ts');
  }

  return RAMLClient;
};
//# sourceMappingURL=index.js.map