const AJAX_TIMEOUT = 60000;

module.exports = ajax;

function ajax(options, next) {
  const { url, proxy } = options;
  const XHR = options.xhr || XMLHttpRequest;
  const method = (options.method || 'GET').toUpperCase();
  // eslint-disable-next-line new-cap
  const xhr = new XHR();
  const async = options.sync !== true;
  const timeout = +options.timeout || AJAX_TIMEOUT;
  let ajaxTimeout;

  /**
   * Wraps callback functions to remove xhr data.
   *
   * @param  {Function} fn
   * @return {Function}
   */
  function complete(fn) {
    return function (...args) {
      window.clearTimeout(ajaxTimeout);

      // Remove references to unused functions.
      xhr.onload = null;
      xhr.onerror = null;
      xhr.onabort = null;

      return fn.apply(this, args);
    };
  }

  xhr.open(method, `${proxy || ''}${url}`, async);

  // Sets all request headers before we make the request.
  Object.keys(options.headers).forEach((key) => {
    xhr.setRequestHeader(key, options.headers[key]);
  });

  // Enable hooking into the ajax request before we send it.
  if (options.beforeSend) {
    options.beforeSend(xhr);
  }

  // Successful callback.
  xhr.onload = complete(() => next(null, buildResponse(xhr)));

  // Failure callback.
  xhr.onabort = complete(() => next(new Error(xhr.statusText || 'AJAX request aborted'), buildResponse(xhr)));

  xhr.onerror = complete(() => next(new Error(xhr.statusText || 'AJAX request errored'), buildResponse(xhr)));

  // Set a request timeout. Modern browsers can set a `timeout` property
  // which works the same, but we'll use a timeout for consistency.
  if (async) {
    ajaxTimeout = window.setTimeout(complete(() => {
      xhr.abort();

      // Calls the `next` function with the timeout details.
      return next(new Error(`AJAX timeout of ${timeout}ms exceeded`), buildResponse(xhr));
    }), timeout);
  }

  xhr.send(options.body);
}

function buildResponse(xhr) {
  const headers = {};

  xhr.getAllResponseHeaders()
    .trim()
    .split('\n')
    .forEach((header) => {
      const index = header.indexOf(':');
      const key = header.slice(0, index).trim().toLowerCase();
      const value = header.slice(index + 1).trim();

      headers[key] = value;
    });

  return {
    body: xhr.responseText,
    statusCode: xhr.status,
    method: xhr.method,
    rawRequest: xhr,
    headers
  };
}
