const template = require('./generator/template');
const triggerAuthentication = require('./authentication');

/**
 * Map authentication types to automatic preference.
 *
 * @type {Array}
 */
const ORDER_PREFERENCE = ['OAuth 2.0', 'OAuth 1.0', 'Basic Authentication'];

/**
 * Required authentication keys used to check the options object.
 *
 * @type {Object}
 */
const DEFAULT_REQUIRED_FIELDS = {
  'OAuth 1.0': {
    consumerKey: true,
    consumerSecret: true
  },
  'OAuth 2.0': {
    clientId: true,
    clientSecret: true
  },
  'Basic Authentication': {
    username: true,
    password: true
  }
};

/**
 * Possible tokens to be filled out by the user.
 *
 * @type {Object}
 */
const DEFAULT_POSSIBLE_FIELDS = {
  'OAuth 1.0': ['consumerKey', 'consumerSecret'],
  'OAuth 2.0': ['clientId', 'clientSecret', 'scopes'],
  'Basic Authentication': ['username', 'password']
};

/**
 * These fields must be hidden in the UI.
 *
 * @type {Object}
 */
const SECRET_FIELDS = {
  clientSecret: true,
  consumerSecret: true,
  password: true
};

/**
 * Returns an object of available keys and whether they are required.
 *
 * @param  {Object} scheme
 * @return {Object}
 */
const requiredTokens = function (scheme) {
  const keys = Object.assign({}, DEFAULT_REQUIRED_FIELDS[scheme.type]);

  // Special case is required for OAuth2 implicit auth flow.
  if (scheme.type === 'OAuth 2.0') {
    keys.clientSecret = !scheme.settings.authorizationGrants.some(grant => grant === 'token' || grant === 'implicit');
  }

  return keys;
};

/**
 * Sanitize scopes to be an array.
 *
 * @param  {*}     scopes
 * @return {Array}
 */
const sanitizeScope = function (scopes) {
  if (typeof scopes === 'string') {
    return scopes.split(' ');
  }

  if (!Array.isArray(scopes)) {
    return [];
  }

  return scopes;
};

/**
 * Prompt the user for authentication tokens based on a scheme. We can pass in
 * the current options object to help decide what data to display to the user.
 *
 * @param {Object}   scheme
 * @param {Object}   options
 * @param {Function} done
 */
const promptTokens = function (scheme, options) {
  const needsTokens = requiredTokens(scheme);

  // Generate an array of the tokens to use with our prompt and filter
  // explicitly not required tokens.
  const possibleTokens = DEFAULT_POSSIBLE_FIELDS[scheme.type].filter(token => needsTokens[token] !== false);

  // Multiple ways of setting the scope option.
  options.scopes = sanitizeScope(options.scope || options.scopes);
  delete options.scope;

  const prompts = {};
  prompts.keys = [];

  possibleTokens.forEach((key) => {
    if (key === 'scopes') {
      const scopes = sanitizeScope(scheme.settings.scopes);

      if (scopes && scopes.length) {
        prompts.scopes = scopes;
        prompts.keys.push(key);
      }
    } else {
      // DEFAULT_FIELD_NAMES[key];
      prompts.keys.push(key);
    }
  });

  if (prompts.keys.every(key => options[key] || (options.credentials && options.credentials[key]))) {
    return Promise.resolve(Object.assign({}, options.credentials, options));
  }

  if (options.prompt) {
    return options.prompt(prompts)
      .then($prompts => Object.assign({}, options.credentials, $prompts, options));
  }

  return Promise.reject(new Error(`Missing required fields: ${prompts.keys.filter(key => !options[key]).join(', ')}`));
};

/**
 * Authenticate using an authentication scheme and passed in options.
 *
 * @param  {Object}   scheme
 * @param  {Object}   options
 * @param  {Function} done
 */
const authenticate = function (scheme, options) {
  const opts = Object.assign({
    type: scheme.type
  }, options, scheme.settings);

  // Interpolate all uri parameters.
  [
    'accessTokenUri',
    'authorizationUri',
    'requestTokenUri',
    'authorizationUri',
    'tokenCredentialsUri'
  ].forEach((param) => {
    opts[param] = template(opts[param], opts.baseUriParameters);
  });

  return triggerAuthentication(opts)
    .then(tokens => ({ scheme, options, tokens }));
};

/**
 * Return the preferred scheme option from an object of every scheme.
 *
 * @param  {Object} schemes
 * @return {Object}
 */
const preferredScheme = function (schemes) {
  // Fix `undefined` schemes issues.
  if (!schemes || !Object.keys(schemes).length) {
    return null;
  }

  const method = ORDER_PREFERENCE.find((preference) => {
    const intersectable = Object.keys(schemes).map(scheme => schemes[scheme].type);

    return intersectable.some(item => item === preference);
  });

  // Return an essentially random but consistent scheme.
  if (!method) {
    return schemes[Object.keys(schemes)[0]];
  }

  // Find the scheme that matched our preferred method.
  return schemes[Object.keys(schemes).find(key => schemes[key].type === method)];
};

/**
 * Retrieve authentication tokens and method any way possible. It will attempt
 * to resolve automatically. If that is not possible, it will defer to
 * prompting the user.
 *
 * @param {Object}   schemes
 * @param {Object}   options
 * @param {Function} done
 */
const retrieveTokens = function (schemes, options) {
  const scheme = preferredScheme(schemes);

  if (!scheme) {
    return Promise.reject(new Error('No schemes available'));
  }

  return promptTokens(scheme, options)
    .then(tokens => ({ scheme, tokens }));
};

/**
 * Attempt to magically resolve to the first working authentication method. If
 * we fail, we need to fall back to manual authentication options with the
 * optimal authentication scheme available.
 *
 * @param {Object}   schemes
 * @param {Object}   options
 * @param {Function} done
 */
const resolveScheme = function (schemes, options) {
  return retrieveTokens(schemes, options)
    .then(({ scheme, tokens }) => authenticate(scheme, tokens));
};

/**
 * Export a function that will contain all the logic for automagically
 * selecting an appropriate authentication method and prompting the user
 * for the following steps.
 *
 * @param {Object}   schemes
 * @param {String}   method
 * @param {Object}   options
 */
module.exports = function (schemes, method, options) {
  const auth = Object.assign({}, options);

  // If no authentication method has been passed in, attempt to pick our own.
  if (!method) {
    return resolveScheme(schemes, auth);
  }

  // Ensure we are attempting to authenticate with a valid method.
  if (!Object.prototype.hasOwnProperty.call(schemes, method)) {
    return Promise.reject(new Error(`The only available authentication methods are: ${
      Object.keys(schemes).map(JSON.stringify).join(', ')}`));
  }

  const scheme = schemes[method];

  // Finally we have everything we need and can initiate authentication.
  return authenticate(scheme, auth);
};

/**
 * Expose secret field names.
 */
module.exports.SECRET_FIELDS = SECRET_FIELDS;
