const qs = require('querystring');
const url = require('url');

const authWindow = require('../authWindow');
const oAuth1Request = require('./request');
const request = require('request');

/**
 * Simple constant for the url encoded content type.
 *
 * @type {String}
 */
const URL_ENCODED = 'application/x-www-form-urlencoded';

/**
 * Send a request to get the initial OAuth request tokens.
 *
 * @param {String}   url
 * @param {Function} done
 */
const getRequestToken = function (options) {
  return new Promise((resolve, reject) => {
    try {
      oAuth1Request({
        url: options.requestTokenUri,
        proxy: options.proxy,
        xhr: options.xhr,
        method: 'POST',
        oauth1: Object.assign({
          oauthCallback: options.redirectUri
        }, options),
        headers: {
          'Content-Type': URL_ENCODED
        }
      }, (err, xhr) => {
        if (err) {
          reject(err);
          return;
        }

        if (xhr.statusCode !== 200) {
          reject(new Error(xhr.body || 'Invalid Request'));
          return;
        }

        const content = qs.parse(xhr.body);

        resolve({
          oauthToken: content.oauth_token,
          oauthTokenSecret: content.oauth_token_secret
        });
      });
    } catch (err) {
      reject(err);
    }
  });
};

/**
 * Get the final access token by passing in the request options and the token
 * verifier.
 *
 * @param {Object}   options
 * @param {String}   verifier
 * @param {Function} done
 */
const getAccessToken = function (options, verifier) {
  return new Promise((resolve, reject) => {
    request({
      url: options.tokenCredentialsUri,
      proxy: options.proxy,
      xhr: options.xhr,
      method: 'POST',
      oauth1: options,
      headers: {
        'Content-Type': URL_ENCODED
      },
      sync: false,
      body: qs.stringify({
        oauth_verifier: verifier
      })
    }, (err, xhr) => {
      if (err) {
        reject(err);
        return;
      }

      try {
        const response = qs.parse(xhr.body);

        const data = {
          response,
          oauthToken: response.oauth_token,
          oauthTokenSecret: response.oauth_token_secret
        };

        // Delete data that has been pulled off the response object to avoid
        // duplication.
        delete response.oauth_token;
        delete response.oauth_token_secret;

        if (!Object.keys(data.response).length) {
          delete data.response;
        }

        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  });
};

module.exports = oAuth1Flow;

/**
 * Trigger the full OAuth1 authentication flow.
 *
 * @param {Object}   options
 * @param {Function} done
 */
function oAuth1Flow(options) {
  if (typeof options.consumerKey !== 'string') {
    return Promise.reject(new TypeError('"consumerKey" expected'));
  }

  if (typeof options.consumerSecret !== 'string') {
    return Promise.reject(new TypeError('"consumerSecret" expected'));
  }

  if (typeof options.requestTokenUri !== 'string') {
    return Promise.reject(new TypeError('"requestTokenUri" expected'));
  }

  if (typeof options.authorizationUri !== 'string') {
    return Promise.reject(new TypeError('"authorizationUri" expected'));
  }

  if (typeof options.tokenCredentialsUri !== 'string') {
    return Promise.reject(new TypeError('"tokenCredentialsUri" expected'));
  }

  return getRequestToken(options)
    .then(data => new Promise((resolve, reject) => {
      const popup = authWindow(`${options.authorizationUri}?${qs.stringify({
        oauth_token: data.oauthToken
      })}`, (err) => {
        if (err) {
          reject(err);
        }

        resolve();
      });

      global.authenticateOAuth = function (href) {
        popup.close();
        delete global.authenticateOAuth;

        if (href.substr(0, options.redirectUri.length) !== options.redirectUri) {
          reject(new Error('Invalid redirect URI'));
          return;
        }

        const response = url.parse(href, true).query;

        if (response.oauth_token !== data.oauthToken) {
          reject(new Error('Invalid OAuth token response'));
          return;
        }

        getAccessToken(Object.assign({
          oauthToken: response.oauth_token
        }, options), response.oauth_verifier)
          .then(accessToken => resolve(accessToken));
      };
    }));
}
