const qs = require('querystring');
const url = require('url');
const createHmac = require('create-hmac');
const request = require('request');
const { isObject } = require('../../utils');
const { encodeData, normalizeUrl } = require('./utils');

module.exports = oAuth1Request;

/**
 * Simple constant for the url encoded content type.
 *
 * @type {String}
 */
const URL_ENCODED = 'application/x-www-form-urlencoded';

/**
 * Check whether the parameters name is a valid OAuth parameter.
 *
 * @param  {String}  param
 * @return {Boolean}
 */
const isParamAnOAuthParameter = function (param) {
  return (/^oauth_/).test(param);
};

/**
 * Transform an object of key, value pairs to an array of arrays.
 *
 * @param  {Object} obj
 * @return {Array}
 */
const paramsToArray = function (obj) {
  return Object.keys(obj).map(key => [key, obj[key]]);
};

/**
 * Generate the Authorization header from an order parameter array.
 *
 * @param  {Object} data
 * @param  {Array}  params
 * @return {String}
 */
const buildAuthorizationHeaders = function (data, params) {
  return `OAuth realm="${normalizeUrl(data.url)}",${
    params
      .filter(param => isParamAnOAuthParameter(param[0]))
      .map(param => `${param[0]}="${param[1]}"`)
      .join(',')
  }`;
};

/**
 * Return the current timestamp in seconds since January 1, 1970 00:00:00 GMT.
 *
 * @return {Number}
 */
const getTimestamp = function () {
  return Math.floor(Date.now() / 1000);
};

/**
 * Generate a random nonce string.
 *
 * @return {String}
 */
const getNonce = function () {
  const chars = [
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9'
  ];

  let nonce = '';

  for (let i = 0; i < 32; i += 1) {
    nonce += chars[Math.floor(Math.random() * chars.length)];
  }

  return nonce;
};

/**
 * Transform an array of parameters (in nested array form) to a query string.
 *
 * @param  {Array}  array
 * @return {String}
 */
const arrayToParams = function (array) {
  return array.map(args => `${args[0]}=${args[1]}`).join('&');
};

/**
 * Sort query string parameters (represented as an array of arrays) by name and
 * then value.
 *
 * @param  {Array} argPairs
 * @return {Array}
 */
const sortRequestParams = function (argPairs) {
  return argPairs.sort((a, b) => {
    if (a[0] === b[0]) {
      return a[1] < b[1] ? -1 : 1;
    }

    return a[0] < b[0] ? -1 : 1;
  });
};

/**
 * Create the base signature string for hashing.
 *
 * @param  {Object} data
 * @return {String}
 */
const createSignatureBase = function (params, data) {
  return [
    data.method.toUpperCase(),
    encodeData(normalizeUrl(data.url)),
    encodeData(arrayToParams(params))
  ].join('&');
};

/**
 * Generate a signature string combining the base signature with consumer
 * secrets.
 *
 * @param  {String} base
 * @param  {Object} data
 * @return {String}
 */
const createSignature = function (base, options) {
  const key = [
    encodeData(options.consumerSecret), encodeData(options.oauthTokenSecret)
  ].join('&');

  let hash = key;

  if (options.signatureMethod === 'HMAC-SHA1') {
    hash = createHmac('sha1', key).update(base).digest('base64');
  }

  return hash;
};

/**
 * Generate a signature from the AJAX data.
 *
 * @param  {Array}  params
 * @param  {Object} data
 * @return {String}
 */
const getSignature = function (params, data) {
  const signatureBase = createSignatureBase(params, data);
  return createSignature(signatureBase, data.oauth1);
};

/**
 * Prepare ordered OAuth parameters based on the provided data.
 *
 * @param  {Object} data
 * @return {Array}
 */
const prepareParameters = function (data) {
  const params = paramsToArray(Object.assign({
    oauth_timestamp: getTimestamp(),
    oauth_nonce: getNonce(),
    oauth_version: '1.0',
    oauth_signature_method: data.oauth1.signatureMethod,
    oauth_consumer_key: data.oauth1.consumerKey
  }, data.url.query));

  // Attach the token query parameter if we have one.
  if (data.oauth1.oauthToken) {
    params.push(['oauth_token', data.oauth1.oauthToken]);
  }

  if (data.oauth1.oauthCallback) {
    params.push(['oauth_callback', data.oauth1.oauthCallback]);
  }

  let contentType = Object.keys(data.headers || {})
    .map(key => [key, data.headers[key]])
    .find(header => header[0].toLowerCase() === 'content-type');

  if (!contentType) {
    contentType = URL_ENCODED;
    data.headers['Content-Type'] = contentType;
  } else {
    // eslint-disable-next-line prefer-destructuring
    contentType = contentType[1];
  }

  if (contentType === URL_ENCODED) {
    if (typeof data.body === 'string') {
      data.body = qs.parse(data.body);
    }

    if (isObject(data.body)) {
      const body = paramsToArray(data.body);
      data.body = arrayToParams(body);
      params.push(...body);
    }
  }

  const sortedParams = sortRequestParams(params);

  sortedParams.push(['oauth_signature', encodeData(getSignature(sortedParams, data))]);

  return sortedParams;
};

/**
 * Allow a new ajax flow for OAuth1-based URLs. Accepts an `oauth1` property
 * on the data object in the format that is returned from the middleware.
 *
 * @param {Object}   data
 * @param {Function} next
 */

function oAuth1Request(data, done) {
  // Check we have an oauth1 object for attempting to mixin keys.
  if (isObject(data.oauth1)) {
    if (!data.oauth1.signatureMethod) {
      data.oauth1.signatureMethod = 'HMAC-SHA1';
    }

    // Parse the url for augmenting the query string parameters. Needed in
    // multiple places throughout the flow, so we can minimize the number of
    // parses by doing it once at the start.
    data.url = url.parse(data.url, true);

    // Delete parameters specific to re-adding the query string, since we need
    // to regenerate the query string without OAuth params.
    delete data.url.href;
    delete data.url.path;
    delete data.url.search;

    const orderedParams = prepareParameters(data);
    const authorization = buildAuthorizationHeaders(data, orderedParams);

    data.headers.Authorization = authorization;

    data.url.query = arrayToParams(paramsToArray(data.url.query).filter(param => !isParamAnOAuthParameter(param[0])));

    // Reattach the query string if we have one available.
    if (data.url.query) {
      data.url.search = `?${data.url.query}`;
      data.url.path = data.url.pathname + data.url.search;
    }

    data.sync = true;
    data.url = url.format(data.url);
  }

  return request(data, done);
}
