const url = require('url');
const qs = require('querystring');
const { omit } = require('../../utils');

module.exports = {
  authResponse,
  erroredResponse,
  handleAjaxResponse,
  sanitizeOptions
};

const JSON_REGEXP = /^application\/([\w!#$%&*`\-.^~]*\+)?json$/i;

/**
 * Handle the xhr response for OAuth requests consistently.
 *
 * @param  {Object}   options
 * @param  {Function} done
 * @return {Function}
 */
function handleAjaxResponse(xhr, options) {
  const { body } = xhr;
  const mime = (xhr.headers['content-type'] || '').split(';')[0];

  if (JSON_REGEXP.test(mime)) {
    return authResponse(options, JSON.parse(body));
  }

  return authResponse(options, qs.parse(body));
}

function sanitizeOptions(options) {
  const oauthCallback = options.oauthCallback || '/authentication/oauth.html';
  const REDIRECT_URI = global.location ? url.resolve(global.location.href, oauthCallback) : null;

  // Extend an default options object.
  const opts = Object.assign({
    redirectUri: REDIRECT_URI
  }, options);

  // Fix up reference to the `scopes` array.
  opts.scope = opts.scope || opts.scopes;

  if (Array.isArray(opts.scope)) {
    opts.scope = opts.scope.join(' ');
  }

  // Remove unused `scopes` property.
  delete opts.scopes;

  return opts;
}

/**
 * Validate an OAuth2 response object.
 *
 * @param {Object}   options
 * @param {Object}   response
 * @param {Function} done
 */
function authResponse(options, response) {
  // Check the response object for errors.
  if (erroredResponse(response)) {
    throw new Error(erroredResponse(response));
  }

  // Filter the response data of known parameters.
  const data = {
    scope: response.scope || options.scope,
    response: omit(
      response,
      'access_token',
      'refresh_token',
      'token_type',
      'expires_in',
      'scope', 'state',
      'error',
      'error_description',
      'error_uri'
    ),
    accessToken: response.access_token
  };

  // Omit the response object altogether if no response is available.
  if (!Object.keys(data.response).length) {
    delete data.response;
  }

  if (response.token_type) {
    data.tokenType = response.token_type;
  }

  if (+response.expires_in) {
    data.expires = Date.now() + (response.expires_in * 1000);
  }

  if (response.refresh_token) {
    data.refreshToken = response.refresh_token;
  }

  return data;
}

/**
 * Format error response types to regular strings for displaying the clients.
 * Reference: http://tools.ietf.org/html/rfc6749#section-4.1.2.1
 *
 * @type {Object}
 */
const ERROR_RESPONSES = {
  invalid_request: [
    'The request is missing a required parameter, includes an',
    'invalid parameter value, includes a parameter more than',
    'once, or is otherwise malformed.'
  ].join(' '),
  invalid_client: [
    'Client authentication failed (e.g., unknown client, no',
    'client authentication included, or unsupported',
    'authentication method).'
  ].join(' '),
  invalid_grant: [
    'The provided authorization grant (e.g., authorization',
    'code, resource owner credentials) or refresh token is',
    'invalid, expired, revoked, does not match the redirection',
    'URI used in the authorization request, or was issued to',
    'another client.'
  ].join(' '),
  unauthorized_client: [
    'The client is not authorized to request an authorization',
    'code using this method.'
  ].join(' '),
  unsupported_grant_type: [
    'The authorization grant type is not supported by the',
    'authorization server.'
  ].join(' '),
  access_denied: [
    'The resource owner or authorization server denied the request.'
  ].join(' '),
  unsupported_response_type: [
    'The authorization server does not support obtaining',
    'an authorization code using this method.'
  ].join(' '),
  invalid_scope: [
    'The requested scope is invalid, unknown, or malformed.'
  ].join(' '),
  server_error: [
    'The authorization server encountered an unexpected',
    'condition that prevented it from fulfilling the request.',
    '(This error code is needed because a 500 Internal Server',
    'Error HTTP status code cannot be returned to the client',
    'via an HTTP redirect.)'
  ].join(' '),
  temporarily_unavailable: [
    'The authorization server is currently unable to handle',
    'the request due to a temporary overloading or maintenance',
    'of the server.'
  ].join(' ')
};

/**
 * Return the formatted error string.
 *
 * @param  {Object} data
 * @return {String}
 */
function erroredResponse(data) {
  return ERROR_RESPONSES[data.error] || data.error || data.error_message;
}
