const qs = require('querystring');
const template = require('./template');
const sanitizeAST = require('./sanitizeAst');
const buildDocumentation = require('./documentation');
const basicAuthRequest = require('../authentication/basic/request');
const oAuth1Request = require('../authentication/oauth1/request');
const oAuth2Request = require('../authentication/oauth2/request');
const ajaxRequest = require('request');
const { isObject } = require('../utils');

const CONFIG_PROPERTY = '!config';
const CLIENT_PROPERTY = '!client';
const JSON_REGEXP = /^application\/([\w!#$%&*`\-.^~]*\+)?json$/i;

const HTTP_METHODS = ['get', 'head', 'put', 'post', 'patch', 'delete'];
const CONFIG_OPTIONS = [
  'body',
  'sync',
  'proxy',
  'xhr',
  'uriParameters',
  'baseUri',
  'baseUriParameters',
  'headers',
  'query',
  'beforeSend'
];
const OVERRIDABLE_CONFIG_OPTIONS = [
  'body',
  'sync',
  'proxy',
  'xhr',
  'baseUri',
  'beforeSend'
];

/**
 * Map the supported auth types to the known triggers.
 *
 * @type {Object}
 */
const authMap = {
  'OAuth 1.0': 'oauth1',
  'OAuth 2.0': 'oauth2',
  'Basic Authentication': 'basicAuth'
};

/**
 * Transform a data object into a form data instance.
 *
 * @param  {Object}   data
 * @return {FormData}
 */
const toFormData = function (data) {
  const form = new FormData();

  // Iterate over every piece of data and append to the form data object.
  if (data) {
    Object.keys(data).forEach((key) => {
      const value = data[key];

      form.append(key, value);
    });
  }

  return form;
};

/**
 * Transform a data object into a String.
 *
 * @param  {Object}   data
 * @return {String}
 */
const toString = function (data) {
  return String(data);
};

/**
 * Map mime types to their parsers.
 *
 * @type {Array}
 */
const parse = [
  [JSON_REGEXP, JSON.parse],
  ['application/x-www-form-urlencoded', qs.parse]
];

/**
 * Map mime types to their serializers.
 *
 * @type {Array}
 */
const serialize = [
  [JSON_REGEXP, JSON.stringify],
  ['application/x-www-form-urlencoded', qs.stringify],
  ['text/html', toString],
  ['multipart/form-data', toFormData]
];

/**
 * Iterate over an array of match and result values, and return the
 * first matching value.
 *
 * @param  {Array}    array
 * @param  {String}   test
 * @return {Function}
 */
const getMatch = function (array, test) {
  const match = array.find((value) => {
    const check = value[0];

    if (check instanceof RegExp) {
      return check.test(test);
    }

    return check === test;
  });

  return match && match[1];
};

/**
 * Sort mimes by preference.
 *
 * @param  {String} mime
 * @return {Number}
 */
const mimePreference = function (mime) {
  return getMatch([
    [JSON_REGEXP, 3],
    ['application/x-www-form-urlencoded', 2],
    ['multipart/form-data', 1]
  ], mime) || 0;
};

/**
 * Check if a method is a query method (not a body as the argument).
 *
 * @param  {String}  method
 * @return {Boolean}
 */
function isQueryMethod(method) {
  return method === 'get' || method === 'head';
}

/**
 * List of all plain HTTP methods in the format from the AST.
 *
 * @type {Object}
 */
const allHttpMethods = HTTP_METHODS.reduce((acum, current) => {
  acum[current] = {
    method: current
  };

  return acum;
}, {});

/**
 * Return the xhr response mime type.
 *
 * @param  {String} contentType
 * @return {String}
 */
const getMime = function (contentType) {
  return contentType == null ? null : String(contentType).split(';')[0];
};

/**
 * Check if an object is a host object and avoid serializing.
 *
 * @param  {Object}  obj
 * @return {Boolean}
 */
const isHost = function (obj) {
  const str = Object.prototype.toString.call(obj);

  switch (str) {
    case '[object File]':
    case '[object Blob]':
    case '[object String]':
    case '[object Number]':
    case '[object Boolean]':
    case '[object FormData]':
      return true;
    default:
      return false;
  }
};

/**
 * Sanitize a specific configuration option.
 *
 * @type {Object}
 */
const sanitizeOption = {
  query(query) {
    if (typeof query === 'string') {
      return qs.parse(query);
    }

    return query;
  }
};

/**
 * Gets a header from the header object.
 *
 * @param  {Object}  headers
 * @param  {String}  header
 * @return {String}
 */
const findHeader = function (headers, header) {
  if (!headers) {
    return null;
  }

  header = header.toLowerCase();

  const key = Object.keys(headers).find(name => name.toLowerCase() === header);

  return headers[key];
};

/**
 * Sanitize the XHR request into the desired format.
 *
 * @param  {XMLHttpRequest} xhr
 * @return {Object}
 */
const sanitizeXHR = function (xhr) {
  if (!xhr) { return xhr; }

  const mime = getMime(xhr.headers['content-type']);
  let body = xhr.body.trim();

  // Automatically parse all response bodies. Only parse when we have response
  // body content. Errors that occur parsing will be propagated to the client.
  body = body ? (getMatch(parse, mime) || (value => value))(body) : undefined;

  return {
    body,
    status: xhr.statusCode,
    headers: xhr.headers
  };
};

/**
 * Returns a function that can be used to make ajax requests.
 *
 * @param  {Object}   nodes
 * @param  {Object}   method
 * @return {Function}
 */
const httpRequest = function (nodes, method) {
  return function (body, config, async) {
    config = config || {};

    // Map configuration options and merge with the passed in object.
    config = CONFIG_OPTIONS.reduce((acum, option) => {
      if (OVERRIDABLE_CONFIG_OPTIONS.indexOf(option) >= 0) {
        acum[option] = config[option] ? config[option] : nodes.config[option];

        return acum;
      }

      const nodeOption = nodes.config[option];
      const configOption = config && config[option];
      const sanitize = sanitizeOption[option] || (value => value);

      acum[option] = Object.assign({}, sanitize(nodeOption), sanitize(configOption));

      return acum;
    }, {});

    let request = ajaxRequest;
    let mime = getMime(findHeader(config.headers, 'Content-Type'));
    const baseUri = template(config.baseUri, config.baseUriParameters);
    let fullUri = `${baseUri.replace(/\/+$/, '')}/${nodes.join('/')}`;

    // GET and HEAD requests accept the query string as the first argument.
    if (isQueryMethod(method.method)) {
      Object.assign(config.query, sanitizeOption.query(body));
      body = null;
    }

    // Set the config object body to the passed in body.
    if (body != null) {
      config.body = body;
    }

    // Append the query string if one is available.
    if (Object.keys(config.query).length) {
      fullUri += `?${qs.stringify(config.query)}`;
    }

    // If we have no accept header set already, default to accepting
    // everything. This is required because Firefox sets the base accept
    // header to essentially be `html/xml`.
    if (!findHeader(config.headers, 'accept')) {
      config.headers.accept = '*/*';
    }

    // If we were passed in data, attempt to sanitize it to the correct type.
    if (!isHost(config.body)) {
      // Set the correct `Content-Type` header, if none exists. Kind of random
      // if more than one exists - I would suggest setting it yourself.
      if (mime == null) {
        // If we have a method body object, sort the method types by most
        // desirable and fallback to a random content type.
        if (isObject(method.body)) {
          mime = Object.keys(method.body).sort(mimePreference).pop();
        }

        // Set the config to the updated mime type header. If none exists, use
        // `application/json` by default.
        mime = mime || 'application/json';
        config.headers['Content-Type'] = mime;
      }

      // Automatically attempt to serialize the body.
      const serializer = getMatch(serialize, mime);

      if (!serializer) {
        throw new TypeError(`Can not serialize content type of "${mime}"`);
      }

      try {
        config.body = serializer(config.body);
      } catch (e) {
        throw new TypeError(`Could not serialize body: ${e.message}`);
      }
    }

    if (async !== undefined) {
      config.sync = !async;
    }

    const options = {
      url: fullUri,
      body: config.body,
      sync: config.sync,
      xhr: config.xhr,
      proxy: config.proxy,
      method: method.method,
      headers: config.headers,
      beforeSend: config.beforeSend
    };

    // Iterate through `securedBy` methods and accept the first one we are
    // already authenticated for.
    const securedBy = Object.keys(method.securedBy || nodes.client.securedBy || {});

    for (let i = 0; i < securedBy.length; i += 1) {
      const key = securedBy[i];
      const scheme = nodes.client.securitySchemes[key];

      // Scheme is not documented in the RAML security schemes.
      if (!scheme) {
        continue;
      }

      const authenticated = nodes.client.authentication[scheme.type];
      const authType = authMap[scheme.type];

      if (authenticated) {
        options[authType] = authenticated;

        if (authType === 'oauth1') {
          request = oAuth1Request;
        }

        if (authType === 'oauth2') {
          request = oAuth2Request;
        }

        if (authType === 'basicAuth') {
          request = basicAuthRequest;
        }

        break;
      }

      continue;
    }

    if (config.sync) {
      // Awkward sync and async code mixing.
      let response;
      let error;

      request(options, (err, xhr) => {
        if (err) {
          error = err;
          return;
        }

        try {
          response = sanitizeXHR(xhr);
        } catch (innerError) {
          error = innerError;
        }
      });

      if (error) {
        throw error;
      }

      return response;
    }

    return new Promise((resolve, reject) => {
      try {
        request(options, (err, xhr) => {
          if (err) {
            reject(err);
            return;
          }

          resolve(xhr);
        });
      } catch (err) {
        reject(err);
      }
    }).then(sanitizeXHR);
  };
};

/**
 * Attaches XHR request methods to the context object for each available method.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} methods
 * @return {Object}
 */
const attachMethods = function (nodes, context, methods) {
  if (!methods) {
    return context;
  }

  // Attach the available methods to the current context.
  Object.keys(methods).forEach((verb) => {
    const method = methods[verb];

    context[verb] = httpRequest(nodes, method);
  });

  return context;
};

/**
 * Attach a special media extension handler.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} resource
 * @return {Object}
 */
const attachMediaTypeExtension = function (nodes, context, resource) {
  /**
   * Push the extension onto the current route and set relevant headers.
   *
   * @param  {String} extension
   * @return {Object}
   */
  context.extension = function (extension) {
    extension = extension == null ? '' : String(extension);

    // Prepend a period to the extension before adding to the route.
    if (extension && extension.charAt(0) !== '.') {
      extension = `.${extension}`;
    }

    const newContext = {};
    const routeNodes = Object.assign([], nodes);

    // Append the extension to the current route.
    routeNodes[routeNodes.length - 1] += extension;

    attachMethods(routeNodes, newContext, resource.methods);
    attachResources(routeNodes, newContext, resource.resources);

    return newContext;
  };

  // If the media type extension is not required, allow direct invocation of
  // the request method.
  if (resource.uriParameters.mediaTypeExtension.required === false) {
    attachMethods(nodes, context, resource.methods);
  }

  // Iterate over the enum options and automatically attach to the context.
  if (resource.uriParameters.mediaTypeExtension.enum) {
    Object.keys(resource.uriParameters.mediaTypeExtension.enum).forEach((key) => {
      let extension = resource.uriParameters.mediaTypeExtension.enum[key];

      if (extension.charAt(0) === '.') {
        extension = extension.substr(1);
      }

      context[extension] = context.extension(extension);
    });
  }

  return context;
};

/**
 * Generate a context or attach methods and resources to an existing context.
 *
 * @param  {Array}   nodes
 * @param  {Object}  resource
 * @param  {Boolean} hasMediaExtension
 * @param  {Object}  [context]
 * @return {Object}
 */
const newContext = function (nodes, resource, hasMediaExtension, context) {
  context = context || {};

  if (hasMediaExtension) {
    attachMediaTypeExtension(nodes, context, resource);
  } else {
    attachMethods(nodes, context, resource.methods);
    attachResources(nodes, context, resource.resources);
  }

  return context;
};

/**
 * Recurses through a resource object in the RAML AST, generating a dynamic
 * DSL that only allows methods that were defined in the RAML spec.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} resources
 * @return {Object}
 */
function attachResources(nodes, context, resources) {
  if (!resources) {
    return context;
  }

  Object.keys(resources).forEach((route) => {
    const resource = resources[route];
    const routeNodes = Object.assign([], nodes);
    let routeName = route;
    const hasMediaExtension = route.substr(-20) === '{mediaTypeExtension}';

    // Ignore media type extensions in route generation.
    if (hasMediaExtension) {
      routeName = route.slice(0, -20);
      route = routeName;
    }

    // Check the route against our valid uri parameters.
    const templateTags = route.match(template.REGEXP);

    // Push the current route into the route array.
    routeNodes.push(route);

    // If we have template tags available, attach a dynamic route.
    if (templateTags) {
      const routeSuffix = templateTags.join('');

      // The route must end with the chained template tags and have no
      // text between tags.
      if (route.substr(-routeSuffix.length) !== routeSuffix) {
        return;
      }

      // If the route is only a template tag with no static text, use the
      // template tag text as the method name.
      if (templateTags.length === 1 && route === templateTags[0]) {
        routeName = templateTags[0].slice(1, -1);
      } else {
        routeName = route.substr(0, route.indexOf('{'));
      }

      // Avoid adding empty route name cases. This can occur when we have
      // multiple tag names and no front text. For example, `{this}{that}`.
      // This could also occur if for some reason we are passing in a route that
      // isn't dynamic.
      if (!routeName) {
        return;
      }

      // Get the ordered tag names for completion.
      const tags = templateTags.map(param => resource.uriParameters[param.slice(1, -1)]);

      // The route is dynamic, so we set the route name to be a function
      // which accepts the template arguments and updates the path fragment.
      // We'll extend any route already at the same namespace so we can do
      // things like use both `/{route}` and `/route`, if needed.
      context[routeName] = Object.assign((...args) => {
        // Map the tags to the arguments or default arguments.
        const parts = tags.map((tag, index) => {
          // Inject enum parameters if there is only one available enum.
          // TODO: When/if we add validation back, have these routes
          // be generated instead of typed out.
          if (args[index] == null && tag.enum && tag.enum.length === 1) {
            return tag.enum[0];
          }

          // Use any passed in argument - even it's falsy.
          if (index in args) {
            return args[index];
          }

          const param = templateTags[index].slice(1, -1);

          // Fallback to injecting the fallback configuration uri parameter.
          return routeNodes.config && routeNodes.config.uriParameters[param];
        });

        // Change the last path fragment to the proper template text.
        routeNodes[routeNodes.length - 1] = template(route, parts);

        return newContext(routeNodes, resource, hasMediaExtension);
      }, context[routeName]);
    }

    // Handle root `/` resources.
    if (!routeName) {
      context = newContext(routeNodes, resource, hasMediaExtension, context);
    } else {
      context[routeName] = newContext(routeNodes, resource, hasMediaExtension, context[routeName]);
    }
  });

  return context;
}

/**
 * Generate the client object from a sanitized AST object.
 *
 * @param  {Object} ast Passed through `sanitizeAST`
 * @param  {Object} config
 * @return {Object}
 */
const generateClient = function (name, ast, config) {
  const { editor } = config;

  // Generate the base uri parameters and defaults.
  const baseUriParameters = Object.assign({ version: ast.version }, config.baseUriParameters);

  // Generate the root node array. Set properties directly on this array to be
  // copied to the next execution part. We have a global configuration object
  // which can be altered externally at any point, as well as when we finally
  // make a request. For this reason, it's important that we use objects which
  // are passed by reference.
  const nodes = Object.assign([], {
    config: Object.assign({
      baseUri: ast.baseUri
    }, config, {
      baseUriParameters
    }),
    client: {
      securedBy: ast.securedBy,
      authentication: {},
      securitySchemes: ast.securitySchemes,
      baseUriParameters: ast.baseUriParameters
    }
  });

  // Throw an error if the baseUri string is missing.
  if (typeof nodes.config.baseUri !== 'string') {
    throw new Error('A "baseUri" string is required');
  }

  /**
   * The root client implementation is simply a function. This allows us to
   * enter a custom path that may not be supported by the DSL and run any
   * method regardless of whether it was defined in the spec.
   *
   * @param  {String} path
   * @param  {Object} context
   * @return {Object}
   */
  const client = function (path, context) {
    const route = template(path || '', context || {}).replace(/^\/+/, '').split('/');

    return attachMethods(Object.assign([], nodes, route), {}, allHttpMethods);
  };

  client[CONFIG_PROPERTY] = nodes.config;
  client[CLIENT_PROPERTY] = nodes.client;
  client.toDescription = () => 'Client has been successfully created';

  attachResources(nodes, client, ast.resources);

  if (editor && !editor.languages.typescript.javascriptDefaults.getExtraLibs()[`${name}-client.d.ts`]) {
    editor.languages.typescript.javascriptDefaults.addExtraLib(
      buildDocumentation(name, ast),
      `${name}-client.d.ts`
    );
  }

  return client;
};

/**
 * Exports the client generator, which accepts the AST of a RAML document.
 *
 * @return {Object} Dynamic object for constructing API requests from the AST.
 */
module.exports = function (name, ast, config) {
  return generateClient(name, sanitizeAST(ast), config);
};
